#include "common.h"

// data:432F 
sbyte bump_col_left_of_wall;
// data:436E
sbyte bump_col_right_of_wall;
// data:405A
sbyte collision_row;
// data:42C2
sbyte prev_collision_row;
// data:4C0A
sbyte right_checked_col;
// data:408A
sbyte left_checked_col;

// data:4C10
sbyte prev_coll_room[10];
// data:4374
sbyte curr_row_coll_room[10];
// data:3D06
sbyte below_row_coll_room[10];
// data:42D2
sbyte above_row_coll_room[10];
// data:5890
byte curr_row_coll_flags[10];
// data:4CEA
byte above_row_coll_flags[10];
// data:4C4C
byte below_row_coll_flags[10];
// data:5BA0
byte prev_coll_flags[10];

// data:4C0C
word coll_tile_left_xpos;
// These two arrays are indexed with the return value of wall_type.
// data:24BA
const sbyte wall_dist_from_left[] = {0, 10, 0, -1, 0, 0};
// data:24C0
const sbyte wall_dist_from_right[] = {0, 0, 10, 13, 0, 0};

// seg004:0004
void __pascal far check_collisions() {
	short column;
	bump_col_left_of_wall = bump_col_right_of_wall = 0xFF;
	if (char_.action == actions_7_turn) return;
	collision_row = char_.curr_row;
	move_coll_to_prev();
	prev_collision_row = collision_row;
	right_checked_col = MIN(get_tile_div_mod_m7(char_x_right_coll) + 2, 11);
	left_checked_col = get_tile_div_mod_m7(char_x_left_coll) - 1;
	get_row_collision_data(collision_row    , curr_row_coll_room, curr_row_coll_flags);
	get_row_collision_data(collision_row + 1, below_row_coll_room, below_row_coll_flags);
	get_row_collision_data(collision_row - 1, above_row_coll_room, above_row_coll_flags);
	for (column = 9; column >= 0; --column) {
		if (curr_row_coll_room[column] >= 0 &&
			prev_coll_room[column] == curr_row_coll_room[column]
		) {
			// char bumps into left of wall
			if (
				(prev_coll_flags[column] & 0x0F) == 0 &&
				(curr_row_coll_flags[column] & 0x0F) != 0
			) {
				bump_col_left_of_wall = column;
			}
			// char bumps into right of wall
			if (
				(prev_coll_flags[column] & 0xF0) == 0 &&
				(curr_row_coll_flags[column] & 0xF0) != 0
			) {
				bump_col_right_of_wall = column;
			}
		}
	}
}

// seg004:00DF
void __pascal far move_coll_to_prev() {
	byte* row_coll_room_ptr;
	byte* row_coll_flags_ptr;
	short column;
	if (collision_row     == prev_collision_row ||
		collision_row + 3 == prev_collision_row ||
		collision_row - 3 == prev_collision_row
	) {
		row_coll_room_ptr = curr_row_coll_room;
		row_coll_flags_ptr = curr_row_coll_flags;
	} else if (
		collision_row + 1 == prev_collision_row ||
		collision_row - 2 == prev_collision_row
	) {
		row_coll_room_ptr = above_row_coll_room;
		row_coll_flags_ptr = above_row_coll_flags;
	} else {
		row_coll_room_ptr = below_row_coll_room;
		row_coll_flags_ptr = below_row_coll_flags;
	}
	for (column = 0; column < 10; ++column) {
		prev_coll_room[column] = row_coll_room_ptr[column];
		prev_coll_flags[column] = row_coll_flags_ptr[column];
		below_row_coll_room[column] = -1;
		above_row_coll_room[column] = -1;
		curr_row_coll_room[column] = -1;
		// bugfix:
		curr_row_coll_flags[column] = 0;
		below_row_coll_flags[column] = 0;
		above_row_coll_flags[column] = 0;
	}
}

// seg004:0185
void __pascal far get_row_collision_data(short row,byte *row_coll_room_ptr,byte *row_coll_flags_ptr) {
	short right_wall_xpos;
	byte curr_flags;
	short room;
	short column;
	short left_wall_xpos;
	room = char_.room;
	coll_tile_left_xpos = x_bump[left_checked_col + 5] + 7;
	for (column = left_checked_col; column <= right_checked_col; ++column) {
		left_wall_xpos = get_left_wall_xpos(room, column, row);
		right_wall_xpos = get_right_wall_xpos(room, column, row);
		// char bumps into left of wall
		curr_flags = (left_wall_xpos < char_x_right_coll) * 0x0F;
		// char bumps into right of wall
		curr_flags |= (right_wall_xpos > char_x_left_coll) * 0xF0;
		row_coll_flags_ptr[tile_col] = curr_flags;
		row_coll_room_ptr[tile_col] = curr_room;
		coll_tile_left_xpos += 14;
	}
}

// seg004:0226
int __pascal far get_left_wall_xpos(int room,int column,int row) {
	short type;
	type = wall_type(get_tile(room, column, row));
	if (type) {
		return wall_dist_from_left[type] + coll_tile_left_xpos;
	} else {
		return 0xFF;
	}
}

// seg004:025F
int __pascal far get_right_wall_xpos(int room,int column,int row) {
	short type;
	type = wall_type(get_tile(room, column, row));
	if (type) {
		return coll_tile_left_xpos - wall_dist_from_right[type] + 13;
	} else {
		return 0;
	}
}

// seg004:029D
void __pascal far check_bumped() {
	if (
		char_.action != actions_2_hang_climb &&
		char_.action != actions_6_hang_straight &&
		// frames 135..149: climb up
		(char_.frame < 135 || char_.frame >= 149)
	) {
		if (bump_col_left_of_wall >= 0) {
			check_bumped_look_right();
		}
#ifndef FIX_TWO_COLL_BUG
		else
#endif
		if (bump_col_right_of_wall >= 0) {
			check_bumped_look_left();
		}
	}
}

// seg004:02D2
void __pascal far check_bumped_look_left() {
	if ((char_.sword == 2 || char_.direction < dir_0_right) && // looking left
		is_obstacle_at_col(bump_col_right_of_wall)
	) {
		bumped(get_right_wall_xpos(curr_room, tile_col, tile_row) - char_x_left_coll, dir_0_right);
	}
}

// seg004:030A
void __pascal far check_bumped_look_right() {
	if ((char_.sword == 2 || char_.direction == dir_0_right) && // looking right
		is_obstacle_at_col(bump_col_left_of_wall)
	) {
		bumped(get_left_wall_xpos(curr_room, tile_col, tile_row) - char_x_right_coll, dir_FF_left);
	}
}

// seg004:0343
int __pascal far is_obstacle_at_col(int tile_col) {
	short tile_row;
	tile_row = char_.curr_row;
	if (tile_row < 0) {
		tile_row += 3;
	}
	if (tile_row >= 3) {
		tile_row -= 3;
	}
	get_tile(curr_row_coll_room[tile_col], tile_col, tile_row);
	return is_obstacle();
}

// seg004:037E
int __pascal far is_obstacle() {
	if (curr_tile2 == tiles_10_potion) {
		return 0;
	} else if (curr_tile2 == tiles_4_gate) {
		if (! can_bump_into_gate()) return 0;
	} else if (curr_tile2 == tiles_18_chomper) {
		// is the chomper closed?
		if (curr_room_modif[curr_tile_pos] != 2) return 0;
	} else if (
		curr_tile2 == tiles_13_mirror &&
		char_.charid == charid_0_kid &&
		char_.frame >= 39 && char_.frame < 44 && // run-jump
		char_.direction < dir_0_right // right-to-left only
	) {
		curr_room_modif[curr_tile_pos] = 0x56; // broken mirror or what?
		jumped_through_mirror = -1;
		return 0;
	}
	coll_tile_left_xpos = xpos_in_curr_room(x_bump[tile_col + 5]) + 7;
	return 1;
}

// seg004:0405
int __pascal far xpos_in_curr_room(int xpos) {
	if (curr_room != drawn_room) {
		if (curr_room == room_L || curr_room == room_BL) {
			xpos -= 140;
		} else if (curr_room == room_R || curr_room == room_BR) {
			xpos += 140;
		}
	}
	return xpos;
}

// seg004:0448
void __pascal far bumped(sbyte delta_x,sbyte push_direction) {
	// frame 177: dead
	if (char_.alive < 0 && char_.frame != 177) {
		char_.x += delta_x;
		if (push_direction < dir_0_right) {
			// pushing left
			if (curr_tile2 == tiles_20_wall) {
				get_tile(curr_room, --tile_col, tile_row);
			}
		} else {
			// pushing right
			if (curr_tile2 == tiles_12_doortop ||
				curr_tile2 == tiles_7_doortop_with_floor ||
				curr_tile2 == tiles_20_wall
			) {
				++tile_col;
				if (curr_room == 0 && tile_col == 10) {
					curr_room = char_.room;
					tile_col = 0;
				}
				get_tile(curr_room, tile_col, tile_row);
			}
		}
		if (tile_is_floor(curr_tile2)) {
			bumped_floor(push_direction);
		} else {
			bumped_fall();
		}
	}
}

// seg004:04E4
void __pascal far bumped_fall() {
	short action;
	action = char_.action;
	char_.x = char_dx_forward(-4);
	if (action == actions_4_in_freefall) {
		char_.fall_x = 0;
	} else {
		seqtbl_offset_char(45); // fall after jump
		play_seq();
	}
	bumped_sound();
}

// seg004:0520
void __pascal far bumped_floor(sbyte push_direction) {
	short frame;
	short seq_index;
	if (char_.sword != 2 && y_land[char_.curr_row + 1] - char_.y >= 15) {
		bumped_fall();
	} else {
		char_.y = y_land[char_.curr_row + 1];
		if (char_.fall_y >= 22) {
			char_.x = char_dx_forward(-5);
		} else {
			char_.fall_y = 0;
			if (char_.alive) {
				if (char_.sword == 2) {
					if (push_direction == char_.direction) {
						seqtbl_offset_char(65); // hit? with sword
						play_seq();
						char_.x = char_dx_forward(1);
						return;
					} else {
						seq_index = 64; // stand with sword
					}
				} else {
					frame = char_.frame;
					if (frame == 24 || frame == 25 ||
						(frame >= 40 && frame < 43) ||
						(frame >= 102 && frame < 107)
					) {
						seq_index = 46; // soft land
					} else {
						seq_index = 47; // bump
					}
				}
				seqtbl_offset_char(seq_index);
				play_seq();
				bumped_sound();
			}
		}
	}
}

// seg004:05F1
void __pascal far bumped_sound() {
	is_guard_notice = 1;
	play_sound(8); // touching a wall
}

// seg004:0601
void __pascal far clear_coll_rooms() {
	memset_near(prev_coll_room, -1, sizeof(prev_coll_room));
	memset_near(curr_row_coll_room, -1, sizeof(curr_row_coll_room));
	memset_near(below_row_coll_room, -1, sizeof(below_row_coll_room));
	memset_near(above_row_coll_room, -1, sizeof(above_row_coll_room));
	// workaround
	memset_near(prev_coll_flags, 0, sizeof(prev_coll_flags));
	memset_near(curr_row_coll_flags, 0, sizeof(curr_row_coll_flags));
	memset_near(below_row_coll_flags, 0, sizeof(below_row_coll_flags));
	memset_near(above_row_coll_flags, 0, sizeof(above_row_coll_flags));
	prev_collision_row = -1;
}

// seg004:0657
int __pascal far can_bump_into_gate() {
	return (curr_room_modif[curr_tile_pos] >> 2) + 6 < char_height;
}

// seg004:067C
int __pascal far get_some_tile() {
	short distance;
	byte tile;
	determine_col();
	load_frame_to_obj();
	set_char_collision();
	tile = get_tile_at_char();
	if (wall_type(tile) != 0) {
		tile_col = char_.curr_col;
		distance = dist_from_wall_forward(tile);
		if (distance >= 0) {
			loc_59DD:
			if (distance < 14) {
				byte_21271 = 1;
			} else {
				byte_21271 = 2;
				distance = 11;
			}
		} else {
			goto loc_59E8;
		}
	} else {
		loc_59E8:
		tile = get_tile_infrontof_char();
		if (tile == tiles_12_doortop && char_.direction >= dir_0_right) {
			loc_59FB:
			byte_21271 = 0;
			distance = distance_to_edge_weight();
		} else {
			if (wall_type(tile) != 0) {
				tile_col = byte_20C62;
				distance = dist_from_wall_forward(tile);
				if (distance >= 0) goto loc_59DD;
			}
			if (tile == tiles_11_loose) goto loc_59FB;
			if (
				tile == tiles_6_closer ||
				tile == tiles_22_sword ||
				tile == tiles_10_potion
			) {
				distance = distance_to_edge_weight();
				if (distance != 0) {
					byte_21271 = 0;
				} else {
					byte_21271 = 2;
					distance = 11;
				}
			} else {
				if (tile_is_floor(tile)) {
					byte_21271 = 2;
					distance = 11;
				} else {
					goto loc_59FB;
				}
			}
		}
	}
	curr_tile2 = tile;
	return distance;
}

// seg004:076B
void __pascal far check_chomped() {
	short tile_col;
	short tile_row;
	tile_row = char_.curr_row;
	for (tile_col = 0; tile_col < 10; ++tile_col) {
		if (curr_row_coll_flags[tile_col] == 0xFF &&
			get_tile(curr_row_coll_room[tile_col], tile_col, tile_row) == tiles_18_chomper &&
			(curr_room_modif[curr_tile_pos] & 0x7F) == 2 // closed chomper
		) {
			chomped();
		}
	}
}

// seg004:07BF
void __pascal far chomped() {
	curr_room_modif[curr_tile_pos] |= 0x80; // put blood
	if (char_.frame != 178 && char_.room == curr_room) {
		char_.x = x_bump[tile_col + 5] + 7;
		char_.x = char_dx_forward(7 - !char_.direction);
		char_.y = y_land[char_.curr_row + 1];
		take_hp(100);
		play_sound(46); // something chopped
		seqtbl_offset_char(54); // chomped
		play_seq();
	}
}

// seg004:0833
void __pascal far check_gate_push() {
	// Closing gate pushes kid
	short frame;
	short var_4;
	frame = char_.frame;
	if (char_.action == actions_7_turn ||
		frame == 15 || // stand
		(frame >= 108 && frame < 111) // crouch
	) {
		get_tile_at_char();
		var_4 = tile_col;
		if ((curr_tile2 == tiles_4_gate ||
			get_tile(curr_room, --tile_col, tile_row) == tiles_4_gate) &&
			(curr_row_coll_flags[tile_col] & prev_coll_flags[tile_col]) == 0xFF && 
			can_bump_into_gate()
		) {
			bumped_sound();
			// push kid left if var_4 <= tile_col, gate at char's tile
			// push kid right if var_4 > tile_col, gate is left from char's tile
			char_.x += 5 - (var_4 <= tile_col) * 10;
		}
	}
}

// seg004:08C3
void __pascal far check_guard_bumped() {
	short var_2;
	if (
		word_1F950 == 0 &&
		char_.action == actions_1_run_jump &&
		char_.alive < 0 &&
		char_.sword >= 2
	) {
		if (
			get_tile_at_char() == tiles_20_wall ||
			curr_tile2 == tiles_7_doortop_with_floor ||
			(curr_tile2 == tiles_4_gate && can_bump_into_gate()) ||
			(char_.direction >= dir_0_right && (
				get_tile(curr_room, --tile_col, tile_row) == tiles_7_doortop_with_floor ||
				(curr_tile2 == tiles_4_gate && can_bump_into_gate())
			))
		) {
			load_frame_to_obj();
			set_char_collision();
			if (is_obstacle()) {
				var_2 = dist_from_wall_behind(curr_tile2);
				if (var_2 < 0 && var_2 > -13) {
					char_.x = char_dx_forward(-var_2);
					seqtbl_offset_char(65); // hit? with sword
					play_seq();
					load_fram_det_col();
				}
			}
		}
	}
}

// seg004:0989
void __pascal far check_chomped_3() {
	get_tile_at_char();
	if ( ! check_chomped2()) {
		get_tile(curr_room, ++tile_col, tile_row);
		check_chomped2();
	}
}

// seg004:09B0
int __pascal far check_chomped2() {
	if (curr_tile2 == tiles_18_chomper &&
		(curr_room_modif[curr_tile_pos] & 0x7F) == 2		
	) {
		coll_tile_left_xpos = x_bump[tile_col + 5] + 7;
		if (get_left_wall_xpos(curr_room, tile_col, tile_row) < char_x_right_coll &&
			get_right_wall_xpos(curr_room, tile_col, tile_row) > char_x_left_coll
		) {
			chomped();
			return 1;
		} else {
			return 0;
		}
	} else {
		return 0;
	}
}

// seg004:0A10
int __pascal far dist_from_wall_forward(char tile) {
	short type;
	if (tile == tiles_4_gate && ! can_bump_into_gate()) {
		return -1;
	} else {
		coll_tile_left_xpos = x_bump[tile_col + 5] + 7;
		type = wall_type(tile);
		if (type == 0) return -1;
		if (char_.direction < dir_0_right) {
			// looking left
			//return wall_dist_from_right[type] + char_x_left_coll - coll_tile_left_xpos - 13;
			return char_x_left_coll - (coll_tile_left_xpos + 13 - wall_dist_from_right[type]);
		} else {
			// looking right
			return wall_dist_from_left[type] + coll_tile_left_xpos - char_x_right_coll;
		}
	}
}

// seg004:0A7B
int __pascal far dist_from_wall_behind(char tile) {
	short type;
	type = wall_type(tile);
	if (type == 0) {
		return 99;
	} else {
		if (char_.direction >= dir_0_right) {
			// looking right
			//return wall_dist_from_right[type] + char_x_left_coll - coll_tile_left_xpos - 13;
			return char_x_left_coll - (coll_tile_left_xpos + 13 - wall_dist_from_right[type]);
		} else {
			// looking left
			return wall_dist_from_left[type] + coll_tile_left_xpos - char_x_right_coll;
		}
	}
}
